package net.sf.itcb.common.server.batch;

import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

import javax.xml.bind.annotation.XmlRootElement;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.oxm.jaxb.Jaxb2Marshaller;
import org.springframework.ws.server.endpoint.annotation.PayloadRoot;
import org.springframework.xml.transform.StringResult;
import org.springframework.xml.transform.StringSource;



import org.apache.commons.io.IOUtils;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Required;
import org.springframework.core.io.UrlResource;

/**
 * Batch mecanism that :
 * <ul>
 *  <li>parse a directory</li>
 *  <li>read all xml files in this directory</li>
 *  <li>analyse which service function to call</li>
 *  <li>call the function</li>
 *  <li>transform response to XML</li>
 *  <li>log xml response in the original file</li>
 *  <li>move original file to a done directory</li>
 * </ul>
 * 
 * Parameters can be injected by spring configuration.
 * This object implements Runnable and can be scheduled by spring task functionalities.  
 *   
 * @author Pierre Le Roux
 */
public class ItcbCommonBatch implements InitializingBean, Runnable {

	protected final Logger log = LoggerFactory.getLogger(getClass());
	
	/**
	 * Error policy for batch : define if process continue when an error is thrown, if we have to stop process or if we move the file in order to avoid it next time
	 * @author Pierre Le Roux
	 */
	protected enum ErrorPolicy {
		CONTINUE,
		STOP,
		MOVE_FILE
	}

	private Object serviceEndpoint;
	
	protected File urlFromDirectory;
	
	protected File urlToDirectory;
	
	protected File urlErrorDirectory;
	
	protected ErrorPolicy onErrorPolicy;
	
	protected Jaxb2Marshaller jaxbMarshaller;
	
	protected Map<XmlRootElement, Method> requestTypeToMethod;
	
	protected BeforeCallInterceptor beforeCallInterceptor;
	

	/**
	 * Set the endpoint which receive request read in xml files
	 * @param serviceEndpoint your module *ServiceServer bean... It should be something like that "yourmoduleEndpoint"
	 */
	@Required
	public void setServiceEndpoint(Object serviceEndpoint) {
		this.serviceEndpoint=serviceEndpoint;
	}
	
	/**
	 * Set the directory in which xml files are retrieved to be read
	 * @param urlFromResource Directory in which xml files are retrieved to be read
	 */
	@Required
	public void setFromResource(UrlResource urlFromResource) throws IOException {
		if(urlFromResource.getFile().isDirectory()) {
			this.urlFromDirectory=urlFromResource.getFile();
		}
		else {
			this.urlFromDirectory=urlFromResource.getFile().getAbsoluteFile();
		}
	}
	
	/**
	 * Set the directory in which xml files are moved after having been processed
	 * @param  urlToResource Directory in which xml files are moved after having been processed
	 */
	@Required
	public void setToResource(UrlResource urlToResource) throws IOException {
		if(urlToResource.getFile().isDirectory()) {
			this.urlToDirectory=urlToResource.getFile();
		}
		else {
			this.urlToDirectory=urlToResource.getFile().getAbsoluteFile();
		}
	}
	
	/**
	 * Set the directory in which xml files are moved if an error occurs during process
	 * @param  urlErrorResource Directory in which xml files are moved if an error occurs during process
	 */
	@Required
	public void setErrorResource(UrlResource urlErrorResource) throws IOException {
		if(urlErrorResource.getFile().isDirectory()) {
			this.urlErrorDirectory=urlErrorResource.getFile();
		}
		else {
			this.urlErrorDirectory=urlErrorResource.getFile().getAbsoluteFile();
		}
	}
	
	/**
	 * Set the error policy
	 * @param errorPolicy error policy
	 */
	@Required
	public void setErrorPolicy(ErrorPolicy errorPolicy) {
		this.onErrorPolicy=errorPolicy;
	}
	
	@Required
	public void setJaxbMarshaller(Jaxb2Marshaller jaxbMarshaller) {
		this.jaxbMarshaller=jaxbMarshaller;
	}
	
	@Override
	public void afterPropertiesSet() throws Exception {
		requestTypeToMethod = new HashMap<XmlRootElement, Method>();
	}
	

	public void run() {
		log.info("Process beginning for xml files in " + urlFromDirectory.getAbsolutePath());
		File[] listFiles = urlFromDirectory.listFiles(new FileFilter() {
			
			@Override
			public boolean accept(File pathname) {
				return pathname.getName().endsWith(".xml");
			}
		});
		int countFilesDone=0;
		int countFilesException=0;
		int countFilesAvoid=0;
		
		if(listFiles != null) {
			for (File file : listFiles) {
				try {
					if(log.isDebugEnabled()) {
						log.debug("Opening " + file.getAbsolutePath());
					}
					FileInputStream fis = new FileInputStream(file);
					String xml = IOUtils.toString(fis);
					fis.close();
					
					Object request = (Object)jaxbMarshaller.unmarshal(new StringSource(xml));
					Class<? extends Object> requestClass = request.getClass();
					Method method = null;
					XmlRootElement xmlRootElement = requestClass.getAnnotation(XmlRootElement.class);
					System.out.println("requestTypeToMethod.getClass()" +requestClass.getCanonicalName());
					System.out.println("xmlRootElement.name() : " + xmlRootElement.name());
					if(requestTypeToMethod.containsKey(xmlRootElement.name())) {
						method= requestTypeToMethod.get(xmlRootElement);
					}
					else {
						for (Method methodLoop : serviceEndpoint.getClass().getMethods()) {
							PayloadRoot payloadRoot= methodLoop.getAnnotation(PayloadRoot.class);
							
							if(payloadRoot != null && payloadRoot.localPart() != null &&  payloadRoot.localPart().equals(xmlRootElement.name())) {
								System.out.println("payloadRoot.localPart() : " + payloadRoot.localPart());
									method=methodLoop;
									requestTypeToMethod.put(xmlRootElement, method);
							}
						}
					}
					
					if(method == null) {
						throw new Exception("No method matches the xml set inside " + file.getAbsolutePath());
					}
					
					if(log.isDebugEnabled()) {
						log.debug("Calling method " + method);
					}
					
					boolean hasToDoIt = true;
					if(beforeCallInterceptor != null) {
						hasToDoIt = beforeCallInterceptor.doBeforeCall(file, method, request);
					}
					
					String returnInformation;
					if(hasToDoIt) {
						Object response = method.invoke(serviceEndpoint, request);
						StringResult responseStringResult = new StringResult();
						jaxbMarshaller.marshal(response,  responseStringResult);
						returnInformation= responseStringResult.toString();
					}
					else {
						returnInformation="No process has been done as beforeCallInterceptor returns that no action has to be done";
					}
					
					FileOutputStream fos = new FileOutputStream(file, true);
					IOUtils.write("\nResponse :\n" + returnInformation, fos);
					IOUtils.closeQuietly(fos);
					File newFile = new File(urlToDirectory, file.getName()+ (hasToDoIt?".done":".avoid")); // +Calendar.getInstance().toString()
					if( ! file.renameTo(newFile)) {
						throw new FatalException("Moving file after process  " + file.getAbsolutePath()); 
					}
					if(hasToDoIt) {
						countFilesDone++;
					}
					else {
						countFilesAvoid++;
					}
				}
				catch(Exception exception) {
					countFilesException++;
					// If we can't move file, we stop the process
					if(exception instanceof FatalException) {
						throw (FatalException)exception;
					}
					log.error("Exception while processing " + file.getAbsolutePath(), exception);
					if(onErrorPolicy.equals(ErrorPolicy.STOP)){
						throw new RuntimeException(exception);
					}
					else if(onErrorPolicy.equals(ErrorPolicy.MOVE_FILE)){
						File newFile = new File(urlErrorDirectory, file.getName());
						file.renameTo(newFile);
						File logFile = new File(urlErrorDirectory, file.getName() + ".log");
						try {
							
							logFile.createNewFile();
							FileOutputStream fos = new FileOutputStream(logFile);
							IOUtils.write(exception.toString(), fos);
							IOUtils.closeQuietly(fos);
						}
						catch(Exception e) {
							log.error("Exception while creating log file " + logFile.getAbsolutePath(), exception);
						}
						
					}
					else {
						// Nothing : we continue
					}
				}
			}
			
		}
		log.info("End of process");
		log.info("Files done count : " + countFilesDone);
		log.info("Files avoid count : " + countFilesAvoid);
		log.info("Files exception count : " + countFilesException);
	}
	
	
	public class FatalException extends RuntimeException {

		private static final long serialVersionUID = 1L;
		
		public FatalException (String message) {
			super(message);
		}
		
		public FatalException (String message, Throwable cause) {
			super(message, cause);
		}
		
	}


	

}
